/* --------------------------------------------------------------
                              pfs_7seg.c

     Softwaremodul zur Ansteuerung des Padauk PFS154 realisierten
     7-Segmentdekoders.

     MCU      : AVR
     Compiler : avr-gcc

     zusaetzliche Hardware:
       PFS154 - 4 Digit 7-Segmentanzeige

     Wichtig: in die Datenleitung zum PFS-Controller muss
     ein Serienwiderstand von 4,7 kOhm geschaltet werden
     weil ansonsten der PFS Unfug macht (warum auch immer)


     21.06.2023   R. Seelig
   -------------------------------------------------------------- */

#include "pfs_7seg.h"

// seg7bmp[16] = alle Segmente aus
// seg7bmp[17] = hochgestelltes o

const uint8_t seg7bmp[] =               \
  {  0x3f,  0x06,  0x5b,  0x4f, 0x66,  0x6d,  0x7d,  0x07,     \
     0x7f,  0x6f,  0x77,  0x7c, 0x39,  0x5e,  0x79,  0x71,
     0x00,  0x63 };

// seg7buffer beinhaltet das Bitmuster der 7-Segmentanzeige, die
// angezeigt werden soll

uint8_t seg7buffer[] = { 0x3f, 0x3f, 0x3f, 0x3f };

/* --------------------------------------------------
                     pfs7seg_sendbmp

     sendet das in seg7buffer enthaltene Bitmuster
     an den PFS154 LED-Anzeigecontroller
   -------------------------------------------------- */
void pfs7seg_sendbmp(void)
{
  uint8_t i, mask, val;

  tx_clr();
  _delay_us(200);

  for (i= 0; i< 32; i++)
  {

    if (!(i % 8))
    {
      mask= 0x80;
      val= seg7buffer[3-(i / 8)];
    }

    if (val & mask)
    {
      tx_set();
      _delay_us(300);
      tx_clr();
      _delay_us(100);
    }
    else
    {
      tx_set();
      _delay_us(100);
      tx_clr();
      _delay_us(300);
    }
    mask= mask >> 1;
  }
  tx_set();
  _delay_ms(3);
}

#if (set4dez_enable == 1)

  /* --------------------------------------------------
                      pfs7seg_set4dez

       beschreibt den Puffer seg7buffer mit den Bit-
       maps der in val uebergebenen dezimalen Zahl und
       sendet den Puffer

       Uebergabe:
         val       : zu sendende dez. Zahl (-999 .. 9999)
         komma_pos : Position an der ein Dezimalpunkt
                     angezeigt werden soll (1 = rechts,
                     3 = links, 0 = kein Dezimalpunkt
     -------------------------------------------------- */
  void pfs7seg_set4dez(int16_t val, uint8_t komma_pos)
  {
    uint8_t  w;

    if ((val> 9999) || (val < -999))
    {
      seg7buffer[3]= 0x79;
      seg7buffer[2]= 0x50;
      seg7buffer[1]= 0x50;
      seg7buffer[0]= 0x00;
    }
    else
    {
      if (val < 0)
      {
        val= -val;
        seg7buffer[3]= 0x40;                // Minuszeichen
      }
      else
      {
        w= val / 1000;
        seg7buffer[3]= seg7bmp[w];
      }
      val= val - ((uint16_t)w * 1000);
      w= val / 100;
      seg7buffer[2]= seg7bmp[w];
      val= val - ((uint16_t)w * 100);
      w= val / 10;
      seg7buffer[1]= seg7bmp[w];
      val= val - ((uint16_t)w * 10);
      w= val;
      seg7buffer[0]= seg7bmp[w];

      switch (komma_pos)
      {
        case 0  : break;
        case 1  : seg7buffer[1] |= 0x80; break;
        case 2  : seg7buffer[2] |= 0x80; break;
        case 3  : seg7buffer[3] |= 0x80; break;
        default : break;
      }
    }
    pfs7seg_sendbmp();
  }

#endif

#if (set2dez_enable == 1)

  /* --------------------------------------------------
                      pfs7seg_set2dez

       beschreibt den Puffer seg7buffer mit den Bit-
       maps der in val uebergebenen 2 stelligen
       dezimalen Zahl und sendet den Puffer

       Uebergabe:
         val  : zu sendende dez. Zahl
         pos  : Position an der die 8-Bit Zahl aus-
                gegeben wird, 3 = links, 2 = Mitte,
                1 = rechts
     -------------------------------------------------- */
  void pfs7seg_set2dez(int8_t val, uint8_t pos)
  {
    uint8_t  w;

    if ((val> 99) || (val < 0))
    {
      seg7buffer[pos]= seg7bmp[0x0f];
      seg7buffer[pos-1]= seg7bmp[0x0f];
    }
    else
    {
      w= val / 10;
      seg7buffer[pos]= seg7bmp[w];
      val= val - ((uint16_t)w * 10);
      w= val;
      seg7buffer[pos-1]= seg7bmp[w];
    }

    pfs7seg_sendbmp();
  }

#endif

#if (sethex_enable == 1)

  /* --------------------------------------------------
                      pfs7seg_sethex

       beschreibt den Puffer seg7buffer mit den Bit-
       maps der in val uebergebenen hexadezimalen Zahl
       und sendet den Puffer

       Uebergabe:
         val       : zu sendende hexadezimale Zahl
     -------------------------------------------------- */
  void pfs7seg_sethex(uint16_t val)
  {
    uint8_t  w;

    seg7buffer[0]= seg7bmp[val & 0x000f];
    val = val >> 4;
    seg7buffer[1]= seg7bmp[val & 0x000f];
    val = val >> 4;
    seg7buffer[2]= seg7bmp[val & 0x000f];
    val = val >> 4;
    seg7buffer[3]= seg7bmp[val & 0x000f];

    pfs7seg_sendbmp();
  }

  /* --------------------------------------------------
                      pfs7seg_init

       initialisiert den Datenpin als Ausgang
     -------------------------------------------------- */
  void pfs7seg_init(void)
  {
    tx_init();
    tx_set();
    _delay_ms(10);
  }


#endif

