/*--------------------------------------------------------
                        set7_4digit.c

    7-Segment 4 Digit Anzeigecontroller v. 2.0

    Empfaengt ueber ein proprietaeres Protokoll auf einer
    Leitung ein 4*8-Bit Muster fuer 4-stellige 7 Seg-
    mentanzeige und zeigt diese an.


     Compiler  : SDCC 4.0.3
     MCU       : PFS154

                 +----------------+
             PB4 | 1   |___|   16 | PB3
             PB5 | 2           15 | PB2
             PB6 | 3           14 | PB1
             PB7 | 4  PFS154   13 | PB0
                  ----------------
             VDD | 5 (1)   (8) 12 | GND
             PA7 | 6 (2)   (7) 11 | PA0
     ICPDA / PA6 | 7 (3)   (6) 10 | PA4
       Vpp / PA5 | 8 (4)   (5)  9 | PA3 / ICPCK
                 +----------------+


               12 11 10  9  8  7
                :  :  :  :  :  :
         +----------------------------+       a
         |  seg3   seg2   seg1  seg0  |      ---
         |   --     --     --    --   |   f | g | b
         |  |  |   |  |   |  |  |  |  |      ---
         |   --     --     --    --   |   e |   | c
         |  |  |   |  |   |  |  |  |  |      ---
         |   --     --     --    --   |       d
         +----------------------------+
                :  :  :  :  :  :
                1  2  3  4  5  6


 Pin-Nr.:  |  1  |  2  |  3  |  4  |  5  |  6  |  7  |  8  |  9  |  10 |  11 |  12 |
 ----------------------------------------------------------------------------------
 Segment   |  e  |  d  |  dp |  c  |  g  | GX0 |  b  | GX1 | GX2 |  f  |  a  | GX3
 ----------------------------------------------------------------------------------
 Portpin   | PB1 | PB0 | PB4 | PA0 | PB3 | PB5 | PA4 | PB6 | PB7 | PB2 | PA3 | PA7


     Protokoll:            [--- insgesamt 32 Datenbits ---]

     ----|                 |----------|           ~~~~    |-----------
     ....|_________________|          |__________ ~~~~____|

         |                                                |
    Start (lo. max. 2 ms)   Hi fuer mehr                Stop
                            als 150us                   Datenframe
                            entspricht
                            logisch 1,
                            kuerzer 150us
                            entspricht
                            logisch 0

     Datentransfer erfolgt mit MSB Digit 3 zuerst:

     seg_dp, seg_g, seg_f, seg_e, seg_d, seg_c, seg_b, seg_a


     19.06.2023        R. Seelig
  -------------------------------------------------------- */

#include <stdint.h>
#include "pdk_init.h"
#include "pfs1xx_gpio.h"
#include "delay.h"


#define com_kathode      1              // 1 => Version fuer gemeinsame Kathode
                                        // 0 => Version fuer gemeinsame Anode


/*--------------------------------------------------------
    Anschlussverbindung der Anzeige mit dem Controller
  -------------------------------------------------------- */

#define sega_init()    PA3_output_init()
#define sega_set()     PA3_set()
#define sega_clr()     PA3_clr()

#define segb_init()    PA4_output_init()
#define segb_set()     PA4_set()
#define segb_clr()     PA4_clr()

#define segc_init()    PA0_output_init()
#define segc_set()     PA0_set()
#define segc_clr()     PA0_clr()

#define segd_init()    PB0_output_init()
#define segd_set()     PB0_set()
#define segd_clr()     PB0_clr()

#define sege_init()    PB1_output_init()
#define sege_set()     PB1_set()
#define sege_clr()     PB1_clr()

#define segf_init()    PB2_output_init()
#define segf_set()     PB2_set()
#define segf_clr()     PB2_clr()

#define segg_init()    PB3_output_init()
#define segg_set()     PB3_set()
#define segg_clr()     PB3_clr()

#define segdp_init()   PB4_output_init()
#define segdp_set()    PB4_set()
#define segdp_clr()    PB4_clr()

#define mpx0_init()    PB5_output_init()
#define mpx0_set()     PB5_clr()
#define mpx0_clr()     PB5_set()

#define mpx1_init()    PB6_output_init()
#define mpx1_set()     PB6_clr()
#define mpx1_clr()     PB6_set()

#define mpx2_init()    PB7_output_init()
#define mpx2_set()     PB7_clr()
#define mpx2_clr()     PB7_set()

#define mpx3_init()    PA7_output_init()
#define mpx3_set()     PA7_clr()
#define mpx3_clr()     PA7_set()

#define seg7_init()    { sega_init(); segb_init(); segc_init();                  \
                         segd_init(); sege_init(); segf_init();                  \
                         segg_init(); segdp_init();                              \
                         mpx0_init(); mpx1_init(); mpx2_init(); mpx3_init(); }

#define rx_init()      PA6_input_init()      //    Anschlusspin des seriellen Eingangs RxD
#define is_rx()        is_PA6()


volatile uint8_t sw_flag = 0;               //     Zaehlt per Interrupt die Anzeigendigits durch


// Buffer, der die Bitmaps der 7-Segmentanzeige beinhaltet
uint8_t seg7buffer[4] = { 0,0,0,0 };


/*--------------------------------------------------------
    seg7_setbmp

    Zeigt Bitmuster einer 7-Segmentanzeige an. Globale
    Variable "com_kathode" wird beruecksichtigt und be-
    stimmt, ob das Bitmuster fuer eine Anzeige mit ge-
    meinsamer Kathode oder gemeinsamer Anode ausgegeben
    wird.
  -------------------------------------------------------- */
void seg7_setbmp(uint8_t bmp)
{
  #if (com_kathode == 1)

    if (bmp & 0x01) sega_set(); else sega_clr();
    if (bmp & 0x02) segb_set(); else segb_clr();
    if (bmp & 0x04) segc_set(); else segc_clr();
    if (bmp & 0x08) segd_set(); else segd_clr();
    if (bmp & 0x10) sege_set(); else sege_clr();
    if (bmp & 0x20) segf_set(); else segf_clr();
    if (bmp & 0x40) segg_set(); else segg_clr();
    if (bmp & 0x80) segdp_set(); else segdp_clr();

  #else

    if (bmp & 0x01) sega_clr(); else sega_set();
    if (bmp & 0x02) segb_clr(); else segb_set();
    if (bmp & 0x04) segc_clr(); else segc_set();
    if (bmp & 0x08) segd_clr(); else segd_set();
    if (bmp & 0x10) sege_clr(); else sege_set();
    if (bmp & 0x20) segf_clr(); else segf_set();
    if (bmp & 0x40) segg_clr(); else segg_set();
    if (bmp & 0x80) segdp_clr(); else segdp_set();

  #endif
}

/* --------------------------------------------------------
                         seg7_mpx

     multiplext die 4-stellige 7-Segmentanzeige.

     Gibt das Bitmuster von seg7buf[0] .. [3] auf der
     Anzeige aus.
   -------------------------------------------------------- */
void seg7_mpx(void)
{

  #if (com_kathode == 1)


    mpx0_clr(); mpx1_clr(); mpx2_clr(); mpx3_clr();

    switch (sw_flag)
    {
      case 0 :
      {
        seg7_setbmp(seg7buffer[0]);
        mpx0_set();
        break;
      }

      case 1 :
      {
        seg7_setbmp(seg7buffer[1]);
        mpx1_set();
        break;
      }

      case 2 :
      {
        seg7_setbmp(seg7buffer[2]);
        mpx2_set();
        break;
      }

      case 3 :
      {
        seg7_setbmp(seg7buffer[3]);
        mpx3_set();
        break;
      }

      default :
      {
        break;
      }
    }

  #else

    mpx0_set(); mpx1_set(); mpx2_set(); mpx3_set();

    switch (sw_flag)
    {
      case 0 :
      {
        seg7_setbmp(seg7buffer[0]);
        mpx0_clr();
        break;
      }

      case 1 :
      {
        seg7_setbmp(seg7buffer[1]);
        mpx1_clr();
        break;
      }

      case 2 :
      {
        seg7_setbmp(seg7buffer[2]);
        mpx2_clr();
        break;
      }

      case 3 :
      {
        seg7_setbmp(seg7buffer[3]);
        mpx3_clr();
        break;
      }

      default :
      {
        break;
      }
    }

  #endif

  sw_flag++;
  sw_flag= sw_flag % 4;
}


/* --------------------------------------------------------
                         seg7_mpx_init

     initialisiert den Timer2 (8-Bit) fuer einen Interrupt-
     intervall und die benoetigten Portanschluesse
     fuer die Anzeige als Ausgaenge
   -------------------------------------------------------- */
void seg7_mpx_init(void)
{
  seg7_init();

  TM2C = (uint8_t)TM2C_CLK_IHRC;
  TM2S = (uint8_t)TM2S_PRESCALE_DIV64 | TM2S_SCALE_DIV4;
  TM2B = 0x80;

  __engint();                   // grundsaetzlich Interrupt zulassen
  INTEN |= INTEN_TM2;           // Timerinterrupt zulassen
}

/* --------------------------------------------------------
                       interrupt

     der Interrupt-Handler. TM2 ruft den Interrupt mit
     ca. 1 kHz auf.

     Hier muss das Multiplexen der Anzeige eingehaengt
     werden.
   -------------------------------------------------------- */
void interrupt(void) __interrupt(0)
{
  // Interruptquelle ist Timer2
  if (INTRQ & INTRQ_TM2)
  {
    seg7_mpx();
    INTRQ &= ~INTRQ_TM2;          // Interruptanforderung quittieren
  }
}

#define timeout     200

/* --------------------------------------------------------
                         getbmp

     liest auf dem Receive-Pin 32 Bit ein, die das 4x8
     Bitmuster auf der Anzeige repraesentieren.

     Wird ein Datentransfer initiert (mit negativer Flanke)
     und es erfolgt innerhalb von 2 ms keine zu lesende
     Datenbits, erfolgt ein Timeout.

     Protokoll:            [--- insgesamt 32 Datenbits ---]

     ----|                 |----------|           ~~~~    |-----------
     ....|_________________|          |__________ ~~~~____|

         |                                                |
    Start (lo. max. 2 ms)   Hi fuer mehr                Stop
                            als 150us                   Datenframe
                            entspricht
                            logisch 1,
                            kuerzer 150us
                            entspricht
                            logisch 0

     Datentransfer erfolgt mit MSB Digit 3 zuerst:

     seg_dp, seg_g, seg_f, seg_e, seg_d, seg_c, seg_b, seg_a

   -------------------------------------------------------- */
void getbmp(void)
{
  uint8_t  i, cx;
  uint32_t inval;
  uint16_t timocnt;

  while (is_rx());                              // auf Start Datenframe warten

  for (cx= 0; cx < 4; cx++)
  {
    inval = 0;
    // 16 Datenbits einlesen
    for (i= 0; i< 8; i++ )
    {
      inval= inval << 1;
      timocnt= 0;
      while ((!is_rx()) && (timocnt < timeout))     // Startflanke detektieren
      {
        _delay_us(10);
        timocnt++;
      }
      if (timocnt== timeout) { goto timeout_err; }

      timocnt= 0;
      while ((is_rx()) && (timocnt < timeout))     // Pulsedauer erfassen
      {
        _delay_us(10);
        timocnt++;
      }

      if (timocnt== timeout) { goto timeout_err; }

      if (timocnt> 15)
      {
        inval |= 0x01;
      }
    }
    seg7buffer[3-cx]= inval;
  }
  timeout_err:
  return;
}


/* --------------------------------------------------
                           main
   -------------------------------------------------- */
void main(void)
{

  seg7_mpx_init();
  rx_init();


  while(1)
  {
    getbmp();
  }
}
